#include <QApplication>
#include <QMainWindow>
#include <QFileDialog>
#include <QMessageBox>
#include <QFileInfo>

#include "main.h"
#include "patcher.h"

MainWindow::MainWindow(QWidget *parent) : QMainWindow(parent)
{
	setupUi(this);

	connect(browseButton, SIGNAL(clicked()), this, SLOT(browse()));
	connect(buttonBox, SIGNAL(clicked(QAbstractButton *)), this, SLOT(done(QAbstractButton *)));
}

void MainWindow::browse()
{
	QString browsePath = QFileDialog::getExistingDirectory(this, tr("Master Levels Directory"));
	dataPath->setText(browsePath);
}

void MainWindow::done(QAbstractButton *button)
{
	QDialogButtonBox::StandardButton sButton = buttonBox->standardButton(button);

	if(sButton == QDialogButtonBox::Close)
		QApplication::exit();
	else
	{
		if(dataPath->text().isEmpty())
		{
			QMessageBox::critical(this, tr("No Data"), tr("No data directory was specified."));
			return;
		}
		QString path = dataPath->text();

		for(unsigned int i = 0;Patcher::filenames[i] != NULL;i++)
		{
			QFileInfo file(path + "/" + Patcher::filenames[i]);
			if(!file.exists())
			{
				QMessageBox::critical(this, tr("File Not Found"), tr("Could not find the following file: ") + Patcher::filenames[i]);
				return;
			}
		}

		QString doomWad;
		if(musicBox->currentIndex() == Patcher::MS_XASERPICKS)
		{
			doomWad = QFileDialog::getOpenFileName(this, tr("Locate doom.wad"), "./doom.wad", tr("Wad Files (*.wad)"));
			if(doomWad.isEmpty())
			{
				QMessageBox::critical(this, tr("File Not Found"), tr("Music selection requires doom.wad, but was not provided."));
				return;
			}
		}

		QString doom2Wad;
		if(musicBox->currentIndex() == Patcher::MS_XASERPICKS || musicBox->currentIndex() == Patcher::MS_BLZUT3PICKS)
		{
			doom2Wad = QFileDialog::getOpenFileName(this, tr("Locate doom2.wad"), "./doom2.wad", tr("Wad Files (*.wad)"));
			if(doom2Wad.isEmpty())
			{
				QMessageBox::critical(this, tr("File Not Found"), tr("Music selection requires doom2.wad, but was not provided."));
				return;
			}
		}

		QString out = QFileDialog::getSaveFileName(this, tr("Output File"), "./master.wad", tr("Wad Files (*.wad)"));
		if(out.isEmpty())
			return;

		QString ret = Patcher::patch(path.toAscii().constData(), out.toAscii().constData(), (Patcher::MapProgression)mapBox->currentIndex(), (Patcher::MusicSelections)musicBox->currentIndex(), doomWad, doom2Wad);
		if(ret.isEmpty())
		{
			QMessageBox::information(this, tr("Done"), tr("The patching process has been completed!"));
			QApplication::exit();
		}
		else
		{
			QMessageBox::critical(this, tr("Error"), ret);
			return;
		}
	}
}

int main(int argc, char* argv[])
{
	QApplication app(argc, argv);

	MainWindow window;
	window.show();

	return app.exec();
}
#ifdef _WIN32
#include <windows.h>
int CALLBACK WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
{
	return main(0, NULL);
}
#endif
