#include <cstdlib>
#include <cstring>
#include <fstream>

#include <QList>

#include "patcher.h"

using namespace Patcher;
using namespace std;

////////////////////////////////////////////////////////////////////////////////

namespace Patcher {

const char* const filenames[] =
{
	"ATTACK.WAD",
	"BLACKTWR.WAD",
	"BLOODSEA.WAD",
	"VIRGIL.WAD",
	"CANYON.WAD",
	"COMBINE.WAD",
	"CATWALK.WAD",
	"FISTULA.WAD",
	"GARRISON.WAD",
	"GERYON.WAD",
	"MANOR.WAD",
	"MEPHISTO.WAD",
	"MINOS.WAD",
	"NESSUS.WAD",
	"PARADOX.WAD",
	"SUBSPACE.WAD",
	"SUBTERRA.WAD",
	"TEETH.WAD",
	"TTRAP.WAD",
	"VESPERAS.WAD",
	NULL
};

static const unsigned char language[] = {
	"[enu default]\n\n"
	"MNU_MASTERLEVELS = \"Master Levels\";\n\n"
	"HUSTR_ATTACK = \"Attack\";\n"
	"HUSTR_BADDREAM = \"Bad Dream\";\n"
	"HUSTR_BLACKTWR = \"Black Tower\";\n"
	"HUSTR_BLOODSEA = \"Bloodsea Keep\";\n"
	"HUSTR_CANYON = \"Canyon\";\n"
	"HUSTR_CATWALK = \"The Catwalk\";\n"
	"HUSTR_COMBINE = \"The Combine\";\n"
	"HUSTR_FISTULA = \"The Fistula\";\n"
	"HUSTR_GARRISON = \"The Garrison\";\n"
	"HUSTR_GERYON = \"Geryon\";\n"
	"HUSTR_MANOR = \"Titan Manor\";\n"
	"HUSTR_MEPHISTO = \"Mephisto's Maosoleum\";\n"
	"HUSTR_MINOS = \"Minos' Judgement\";\n"
	"HUSTR_NESSUS = \"Nessus\";\n"
	"HUSTR_PARADOX = \"Paradox\";\n"
	"HUSTR_SUBSPACE = \"Subspace\";\n"
	"HUSTR_SUBTERRA = \"Subterra\";\n"
	"HUSTR_TEETH = \"Express Elevator to Hell\";\n"
	"HUSTR_TTRAP = \"Trapped on Titan\";\n"
	"HUSTR_VESPERAS = \"Vesperas\";\n"
	"HUSTR_VIRGIL = \"Virgil's Lead\";\n\n"
	"MASTERLEVELS_CLUSTER_TEXT =\n"
	"	\"You have assaulted and triumphed over the\\n\"\n"
	"	\"most vicious realms that the demented\\n\"\n"
	"	\"minds of our designers could devise. The\\n\"\n"
	"	\"havoc you left behind you as you smashed\\n\"\n"
	"	\"your way through the master levels is\\n\"\n"
	"	\"mute tribute to your prowess.\\n\\n\"\n"
	"	\"You have earned the title of master of\\n\"\n"
	"	\"destruction.\";\n"
};

static const char clusterinfo[] = {
	"clusterdef 11\n"
	"{\n"
	"	flat = \"RROCK17\"\n"
	"	music = \"$MUSIC_READ_M\"\n"
	"	exittext = lookup, \"MASTERLEVELS_CLUSTER_TEXT\"\n"
	"}"
};
static const char xaserinfo[] = {
	"cluster 11\n"
	"{\n"
	"		flat = \"FLOOR7_2\"\n"
	"		music = \"$MUSIC_READ_M\"\n"
	"		exittext = \"BLOODSOAKED AND ADRENALINE FUELED, YOU\",\n"
	"		\"LAUGH MANIACALLY IN THE WAKE OF\",\n"
	"		\"DESTRUCTION. THE DEMONS MAY HAVE BROUGHT\",\n"
	"		\"HELL BACK WITH THEM, BUT YOU CRUSHED THEIR\",\n"
	"		\"EFFORTS AND SHOVED HELL BACK DOWN THEIR\",\n"
	"		\"THROATS. EVEN THE COLOSSAL MIGHT OF THE\",\n"
	"		\"BLACK TOWER COULD NOT STOP YOUR ADVANCE.\",\n"
	"		\"\",\n"
	"		\"BUT STILL, THERE ARE MORE TERRORS TO COME,\",\n"
	"		\"FROM THE SHORES OF TITAN TO THE BOWELS\",      \n"
	"		\"OF HELL ITSELF. AS YOU DIVE INTO THE\",\n"
	"		\"FIERY ABYSS, A SINGLE WORD ECHOES THROUGH\",\n"
	"		\"YOUR HEAD.\",\n"
	"		\"\",\n"
	"		\"PAYBACK.\"\n"
	"}\n\n"
	"cluster 12\n"
	"{\n"
	"		flat = \"BLOOD1\"\n"
	"		music = \"$MUSIC_READ_M\"\n"
	"		exittext = \"WITH A THUNDEROUS CRASH, THE SHATTERED\",\n"
	"		\"REMAINS OF THE ANCIENT BEAST CRUMBLE TO\",\n"
	"		\"PIECES, THE WEIGHT OF THE COLLAPSING\",\n"
	"		\"MAUSOLEUM CRUSHING AND BURYING THE FIEND\",\n"
	"		\"PAST HELL AND INTO OBLIVION.\",\n"
	"		\"\",\n"
	"		\"YOU'VE DONE IT! THE DEMONIC PRESENCE IS NO\",\n"
	"		\"MORE. HAVING CLEANSED HELL ITSELF FROM ITS\",\n"
	"		\"OWN FILTH, YOU HAVE PROCLAIMED YOUR\",\n"
	"		\"VICTORY ACROSS A THOUSAND WORLDS.\",\n"
	"		\"EVEN THE MASTERS OF YOUR OWN DOOM WERE\",\n"
	"		\"NO MATCH FOR YOUR DEVILISH TALENTS IN\",\n"
	"		\"EXTERMINATION. HOW ABOUT IT, MARINE?\",\n"
	"		\"WHAT DOES IT FEEL LIKE TO HAVE TRIUMPHED\",\n"
	"		\"OVER THE IMPOSSIBLE TIME AND TIME AGAIN?\",\n"
	"		\"\",\n"
	"		\"THIS MUST BE WHAT IT FEELS LIKE TO\",\n"
	"		\"BE A GOD. A GOD OF DOOM.\"\n"
	"}\n\n"
	"cluster 13\n"
	"{\n"
	"		flat = \"CEIL4_1\"\n"
	"		music = \"$MUSIC_READ_M\"\n"
	"		entertext =     \"A DISTURBING SILENCE WASHES OVER YOU AS\",\n"
	"		\"YOU ENTER THIS NEW REALM. YOU CAN'T HELP\",\n"
	"		\"BUT FEEL AS IF THIS PLACE WAS NOT\",\n"
	"		\"MEANT TO BE DISTURBED...\",\n"
	"		\"\",\n"
	"		\"TIME TO WAKE UP...\"\n"
	"}"
};
static const char mapinfo[] = {
	"// Master Levels for Doom 2\n\n"
	"episode \"ATTACK\"\n"
	"{\n"
	"	key = \"M\"\n"
	"	name = \"$MNU_MASTERLEVELS\"\n"
	"}\n\n"
	"%47\n\n"
	"map \"ATTACK\" lookup \"HUSTR_ATTACK\"\n"
	"{\n"
	"	next = \"%1\"\n"
	"	secretnext = \"%1\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%23\"\n"
	"}\n\n"
	"map \"BLACKTWR\" lookup \"HUSTR_BLACKTWR\"\n"
	"{\n"
	"	next = \"%2\"\n"
	"	secretnext = \"%2\"\n"
	"	sky1 = \"SKY3\", 0\n"
	"	cluster = %44\n"
	"	music = \"%24\"\n"
	"}\n\n"
	"map \"BLOODSEA\" lookup \"HUSTR_BLOODSEA\"\n"
	"{\n"
	"	next = \"%3\"\n"
	"	secretnext = \"%3\"\n"
	"	sky1 = \"%48\", 0\n"
	"	cluster = %45\n"
	"	map07special\n"
	"	music = \"%25\"\n"
	"}\n\n"
	"map \"CANYON\" lookup \"HUSTR_CANYON\"\n"
	"{\n"
	"	next = \"%4\"\n"
	"	secretnext = \"%4\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%26\"\n"
	"}\n\n"
	"map \"CATWALK\" lookup \"HUSTR_CATWALK\"\n"
	"{\n"
	"	next = \"%5\"\n"
	"	secretnext = \"%5\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%27\"\n"
	"}\n\n"
	"map \"COMBINE\" lookup \"HUSTR_COMBINE\"\n"
	"{\n"
	"	next = \"%6\"\n"
	"	secretnext = \"%6\"\n"
	"	sky1 = \"MSKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%28\"\n"
	"}\n\n"
	"map \"FISTULA\" lookup \"HUSTR_FISTULA\"\n"
	"{\n"
	"	next = \"%7\"\n"
	"	secretnext = \"%7\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%29\"\n"
	"}\n\n"
	"map \"GARRISON\" lookup \"HUSTR_GARRISON\"\n"
	"{\n"
	"	next = \"%8\"\n"
	"	secretnext = \"%8\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%30\"\n"
	"}\n\n"
	"map \"GERYON\" lookup \"HUSTR_GERYON\"\n"
	"{\n"
	"	next = \"%9\"\n"
	"	secretnext = \"%9\"\n"
	"	sky1 = \"MSKY2\", 0\n"
	"	cluster = %45\n"
	"	music = \"%31\"\n"
	"}\n\n"
	"map \"MANOR\" lookup \"HUSTR_MANOR\"\n"
	"{\n"
	"	next = \"%10\"\n"
	"	secretnext = \"%10\"\n"
	"	sky1 = \"MSKY3\", 0\n"
	"	cluster = %45\n"
	"	music = \"%32\"\n"
	"}\n\n"
	"map \"MEPHISTO\" lookup \"HUSTR_MEPHISTO\"\n"
	"{\n"
	"	next = \"%11\"\n"
	"	secretnext = \"%11\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %45\n"
	"	map07special\n"
	"	music = \"%33\"\n"
	"}\n\n"
	"map \"MINOS\" lookup \"HUSTR_MINOS\"\n"
	"{\n"
	"	next = \"%12\"\n"
	"	secretnext = \"%12\"\n"
	"	sky1 = \"MSKY2\", 0\n"
	"	cluster = %45\n"
	"	music = \"%34\"\n"
	"}\n\n"
	"map \"NESSUS\" lookup \"HUSTR_NESSUS\"\n"
	"{\n"
	"	next = \"%13\"\n"
	"	secretnext = \"%13\"\n"
	"	sky1 = \"MSKY2\", 0\n"
	"	cluster = %45\n"
	"	map07special\n"
	"	music = \"%35\"\n"
	"}\n\n"
	"map \"PARADOX\" lookup \"HUSTR_PARADOX\"\n"
	"{\n"
	"	next = \"%14\"\n"
	"	secretnext = \"%14\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%36\"\n"
	"}\n\n"
	"map \"SUBSPACE\" lookup \"HUSTR_SUBSPACE\"\n"
	"{\n"
	"	next = \"%15\"\n"
	"	secretnext = \"%15\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%37\"\n"
	"}\n\n"
	"map \"SUBTERRA\" lookup \"HUSTR_SUBTERRA\"\n"
	"{\n"
	"	next = \"%16\"\n"
	"	secretnext = \"%16\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %44\n"
	"	music = \"%38\"\n"
	"}\n\n"
	"map \"TEETH\" lookup \"HUSTR_TEETH\"\n"
	"{\n"
	"	next = \"%17\"\n"
	"	secretnext = \"%18\"\n"
	"	sky1 = \"SKY1\", 0\n"
	"	cluster = %45\n"
	"	music = \"%39\"\n"
	"}\n\n"
	"map \"BADDREAM\" lookup \"HUSTR_BADDREAM\"\n"
	"{\n"
	"	next = \"%19\"\n"
	"	secretnext = \"%19\"\n"
	"	sky1 = \"SKY3\", 0\n"
	"	cluster = %46\n"
	"	music = \"%40\"\n"
	"}\n\n"
	"map \"TTRAP\" lookup \"HUSTR_TTRAP\"\n"
	"{\n"
	"	next = \"%20\"\n"
	"	secretnext = \"%20\"\n"
	"	sky1 = \"MSKY3\", 0\n"
	"	cluster = %45\n"
	"	music = \"%41\"\n"
	"}\n\n"
	"map \"VESPERAS\" lookup \"HUSTR_VESPERAS\"\n"
	"{\n"
	"	next = \"%21\"\n"
	"	secretnext = \"%21\"\n"
	"	sky1 = \"MSKY2\", 0\n"
	"	cluster = %45\n"
	"	music = \"%42\"\n"
	"}\n\n"
	"map \"VIRGIL\" lookup \"HUSTR_VIRGIL\"\n"
	"{\n"
	"	next = \"%22\"\n"
	"	secretnext = \"%22\"\n"
	"	sky1 = \"MSKY2\", 0\n"
	"	cluster = %45\n"
	"	music = \"%43\"\n"
	"}\n"
};

static const char* const MapProgressions[8][22] =
{
	// Doom Classic (PSN) w/ Secret exit
	{
		"CANYON", "TEETH", "MEPHISTO", "CATWALK", "COMBINE",
		"FISTULA", "GARRISON", "MANOR", "VESPERAS", "PARADOX",
		"NESSUS", "BLOODSEA", "GERYON", "SUBSPACE", "SUBTERRA",
		"TTRAP", "ENDGAME1", "BADDREAM", "ENDGAME1", "VIRGIL",
		"BLACKTWR", "MINOS"
	},
	// Doom Classic (PSN) Strict
	{
		"CANYON", "TEETH", "MEPHISTO", "CATWALK", "COMBINE",
		"FISTULA", "GARRISON", "MANOR", "VESPERAS", "PARADOX",
		"NESSUS", "BLOODSEA", "GERYON", "SUBSPACE", "SUBTERRA",
		"TTRAP", "BADDREAM", "BADDREAM", "ENDGAME1", "VIRGIL",
		"BLACKTWR", "MINOS"
	},
	// Playstation extended w/ Secret exit
	{
		"BLACKTWR", "BLOODSEA", "VIRGIL", "COMBINE", "FISTULA",
		"CATWALK", "GARRISON", "GERYON", "MANOR", "MEPHISTO",
		"MINOS", "NESSUS", "PARADOX", "SUBSPACE", "SUBTERRA",
		"TEETH", "TTRAP", "BADDREAM", "TTRAP", "VESPERAS",
		"ENDGAME1", "CANYON"
	},
	// Xbox extended w/ Secret exit
	{
		"BLACKTWR", "BLOODSEA", "CANYON", "CATWALK", "COMBINE",
		"FISTULA", "GARRISON", "GERYON", "MANOR", "MEPHISTO",
		"MINOS", "NESSUS", "PARADOX", "SUBSPACE", "SUBTERRA",
		"TEETH", "TTRAP", "BADDREAM", "TTRAP", "VESPERAS",
		"VIRGIL", "ENDGAME1"
	},
	// Xbox Strict
	{
		"BLACKTWR", "BLOODSEA", "CANYON", "CATWALK", "COMBINE",
		"FISTULA", "GARRISON", "GERYON", "MANOR", "NESSUS",
		"NESSUS", "NESSUS", "SUBSPACE", "SUBSPACE", "SUBTERRA",
		"TEETH", "BADDREAM", "BADDREAM", "TTRAP", "VESPERAS",
		"VIRGIL", "ENDGAME1"
	},
	// Playstation Strict
	{
		"VIRGIL", "CANYON", "CANYON", "COMBINE", "FISTULA",
		"CATWALK", "GERYON", "GERYON", "MINOS", "MINOS",
		"MINOS", "NESSUS", "PARADOX", "SUBSPACE", "SUBTERRA",
		"VESPERAS", "VESPERAS", "BADDREAM", "VESPERAS", "VESPERAS",
		"ENDGAME1", "CANYON"
	},
	// Xaser
	{
		"CANYON", "VIRGIL", "MEPHISTO", "CATWALK", "FISTULA",
		"SUBSPACE", "COMBINE", "BLACKTWR", "VESPERAS", "TTRAP",
		"ENDGAME1", "NESSUS", "GERYON", "SUBTERRA", "PARADOX",
		"GARRISON", "BLOODSEA", "BADDREAM", "BLOODSEA", "TEETH",
		"MANOR", "MINOS"
	},
	// Xaser w/ Custom text
	{
		"CANYON", "VIRGIL", "MEPHISTO", "CATWALK", "FISTULA",
		"SUBSPACE", "COMBINE", "BLACKTWR", "VESPERAS", "TTRAP",
		"ENDGAME1", "NESSUS", "GERYON", "SUBTERRA", "PARADOX",
		"GARRISON", "BLOODSEA", "BADDREAM", "BLOODSEA", "TEETH",
		"MANOR", "MINOS"
	}
};

static const char* const MapMusic[4][21] =
{
	{
		"D_RUNNIN", "D_SHAWN2", "D_DDTBL2", "D_STALKS", "D_COUNTD",
		"D_BETWEE", "D_DOOM", "D_THE_DA", "D_STLKS3", "D_SHAWN",
		"D_RUNNI2", "D_DOOM2", "D_DEAD2", "D_DDTBLU", "D_IN_CIT",
		"D_DEAD", "D_MESSAG", "D_COUNT2", "D_STLKS2", "D_ROMERO",
		"D_THEDA2"
	},
	{
		"D_RUNNIN", "D_ADRIAN", "D_SHAWN", "D_RUNNIN", "D_RUNNIN",
		"D_RUNNIN", "D_RUNNIN", "D_RUNNIN", "D_DDTBLU", "D_RUNNIN",
		"D_SHAWN", "D_DOOM", "D_SHAWN", "D_RUNNIN", "D_RUNNIN",
		"D_RUNNIN", "D_EVIL", "D_ULTIMA", "D_RUNNIN", "D_IN_CIT",
		"D_COUNTD"
	},
	{
		"D_BETWEE", "D_ADRIAN", "D_SHAWN", "D_THE_DA", "D_DDTBLU",
		"D_STALKS", "D_IN_CIT", "D_DEAD", "D_DDTBLU", "D_ADRIAN",
		"D_SHAWN", "D_DOOM", "D_SHAWN", "D_INTDM2", "D_STALKS",
		"D_DDTBLU", "D_EVIL", "D_ULTIMA", "D_DEAD", "D_IN_CIT",
		"D_COUNTD"
	},
	{
		"D_INTDM2", "D_ADRIAN", "D_E1M7", "D_E2M2", "D_THE_DA",
		"D_DEAD", "D_BETWEE", "D_E3M3", "D_E1M5", "D_MESSAG",
		"D_E1M8", "D_E2M6", "D_ROMERO", "D_IN_CIT", "D_STALKS",
		"D_DDTBLU", "D_E3M8", "D_OPENIN", "D_SHAWN", "D_TENSE",
		"D_README"
	}
};

}

////////////////////////////////////////////////////////////////////////////////

typedef unsigned int UInt32;
typedef unsigned char UInt8;
#define READINT32(pointer) ((UInt32((UInt8)(*pointer))) | (UInt32(UInt8(*(pointer+1)))<<8) | (UInt32(UInt8(*(pointer+2)))<<16) | (UInt32(UInt8(*(pointer+3)))<<24))
#define WRITEINT32_DIRECT(integer) (UInt8)(integer&0xFF),(UInt8)((integer>>8)&0xFF),(UInt8)((integer>>16)&0xFF),(UInt8)((integer>>24)&0xFF)
#define WRITEINT32(pointer, integer) *pointer = (UInt8)(integer&0xFF);*(pointer+1) = (UInt8)((integer>>8)&0xFF);*(pointer+2) = (UInt8)((integer>>16)&0xFF);*(pointer+3) = (UInt8)((integer>>24)&0xFF);

class WadFile
{
	private:
		struct Lump
		{
			public:
				Lump() : size(0), offset(0), data(NULL)
				{
				}

				~Lump()
				{
					if(data != NULL)
						delete[] data;
				}

				char name[9];
				unsigned int size;
				unsigned int offset;
				unsigned char* data;
		};

		QList<Lump *> lumps;

	public:
		~WadFile()
		{
			for(unsigned int i = 0;i < lumps.size();++i)
				delete lumps[i];
		}

/*		void Dump()
		{
			for(vector<Lump *>::iterator lump = lumps.begin();lump != lumps.end();lump++)
				printf("%s\n", (*lump)->name);
		}*/

		void AddLump(const char* name, unsigned int size, const unsigned char* data)
		{
			Lump *lump = new Lump();
			lump->size = size;
			strncpy(lump->name, name, 9);
			lump->data = new unsigned char[size];
			memcpy(lump->data, data, size);
			lumps.append(lump);
		}

		Lump *GetLump(const char* name, unsigned int instance=1)
		{
			for(unsigned int i = 0;i < lumps.size();++i)
			{
				if(strcmp(lumps[i]->name, name) == 0)
				{
					instance--;
					if(instance == 0)
						return lumps[i];
				}
			}
			return NULL;
		}

		void RenameLump(const char* oldName, const char* newName)
		{
			Lump *lump = GetLump(oldName);
			if(lump != NULL)
				strncpy(lump->name, newName, 8);
		}

		bool Save(const char* filename)
		{
			unsigned int dirOffset = 12;
			for(unsigned int i = 0;i < lumps.size();++i)
			{
				lumps[i]->offset = dirOffset;
				dirOffset += lumps[i]->size;
			}

			ofstream stream(filename, ios_base::out | ios_base::binary | ios_base::trunc);
			if(!stream.is_open())
				return false;

			unsigned int numLumps = lumps.size();
			char header[12] = { 'P', 'W', 'A', 'D', WRITEINT32_DIRECT(numLumps), WRITEINT32_DIRECT(dirOffset) };
			if(stream.write(header, 12).fail())
			{
				stream.close();
				return false;
			}

			char* directory = new char[16*numLumps];
			memset(directory, 0, 16*numLumps);
			for(unsigned int i = 0;i < numLumps;i++)
			{
				if(stream.write(reinterpret_cast<char*> (lumps[i]->data), lumps[i]->size).fail())
				{
					stream.close();
					delete[] directory;
					return false;
				}
				WRITEINT32(&directory[i*16], lumps[i]->offset);
				WRITEINT32(&directory[(i*16)+4], lumps[i]->size);
				memcpy(&directory[(i*16)+8], lumps[i]->name, strlen(lumps[i]->name));
			}
			if(stream.write(directory, 16*numLumps).fail())
			{
				stream.close();
				delete[] directory;
				return false;
			}
			stream.close();
			delete[] directory;
			return true;
		}

		void operator<< (const Lump *lump)
		{
			if(lump != NULL)
				AddLump(lump->name, lump->size, lump->data);
		}

		static WadFile *FromFile(const char* in)
		{
			//printf("Opening %s\n", in);
			ifstream stream(in, ios_base::in | ios_base::binary);
			if(!stream.is_open())
				return NULL;

			//printf("Reading header\n");
			char header[12];
			stream.read(header, 12);
			if(stream.fail() || (header[0] != 'P' && header[0] != 'I') || header[1] != 'W' || header[2] != 'A' || header[3] != 'D')
			{
				stream.close();
				return NULL;
			}

			unsigned int numLumps = READINT32(&header[4]);
			unsigned int dirOffset = READINT32(&header[8]);

			//printf("Reading directory\n");
			if(stream.seekg(dirOffset, ios_base::beg).fail())
			{
				stream.close();
				return NULL;
			}
			char* directoryData = new char[numLumps*16];
			if(stream.read(directoryData, numLumps*16).fail())
			{
				delete[] directoryData;
				stream.close();
				return NULL;
			}

			// Read the lumps
			WadFile *wFile = new WadFile();
			for(unsigned int i = 0;i < numLumps;i++)
			{
				if(stream.seekg(READINT32(&directoryData[i*16])).fail())
				{
					delete wFile;
					delete[] directoryData;
					stream.close();
					return NULL;
				}

				unsigned char* data = new unsigned char[READINT32(&directoryData[(i*16)+4])];
				if(stream.read(reinterpret_cast<char*> (data), READINT32(&directoryData[(i*16)+4])).fail())
				{
					delete wFile;
					delete[] directoryData;
					delete[] data;
					stream.close();
					return NULL;
				}
				char name[9];
				strncpy(name, &directoryData[(i*16)+8], 8);
				name[8] = 0;
				wFile->AddLump(name, READINT32(&directoryData[(i*16)+4]), data);
				delete[] data;
			}
			delete[] directoryData;
			stream.close();
			return wFile;
		}
};

QString Patcher::patch(const char* in, const char* out, Patcher::MapProgression progression, Patcher::MusicSelections musicSelections, const QString &doomWad, const QString &doom2Wad)
{
	const char* const * files = filenames;
	WadFile skys; // we'll merge this later
	WadFile output;
	for(unsigned int i = 0;i < NUM_FILES;i++)
	{
		QString path = QString(in) + "/" + files[i];
		WadFile *file = WadFile::FromFile(path.toAscii());
		if(file == NULL)
			return QString("Could not read \"") + files[i] + "\"";

		// Do stuff to the file
		QString mapName = QString(files[i]);
		mapName = mapName.left(mapName.indexOf('.'));
		const char* oldName;
		switch(i)
		{
			case VIRGIL:
				oldName = "MAP03";
				break;
			case MINOS:
				oldName = "MAP05";
				break;
			case BLOODSEA:
			case MEPHISTO:
			case NESSUS:
				oldName = "MAP07";
				break;
			case GERYON:
				oldName = "MAP08";
				break;
			case VESPERAS:
				oldName = "MAP09";
				break;
			case BLACKTWR:
				oldName = "MAP25";
				break;
			case TEETH:
				oldName = "MAP31";
				file->RenameLump("MAP32", "BADDREAM");
				break;
			default:
				oldName = "MAP01";
				break;
		}
		file->RenameLump(oldName, mapName.toAscii());
		output << file->GetLump(mapName.toAscii());
		output << file->GetLump("THINGS");
		output << file->GetLump("LINEDEFS");
		output << file->GetLump("SIDEDEFS");
		output << file->GetLump("VERTEXES");
		output << file->GetLump("SEGS");
		output << file->GetLump("SSECTORS");
		output << file->GetLump("NODES");
		output << file->GetLump("SECTORS");
		output << file->GetLump("REJECT");
		output << file->GetLump("BLOCKMAP");
		if(i == TEETH)
		{
			output << file->GetLump("BADDREAM");
			output << file->GetLump("THINGS",2);
			output << file->GetLump("LINEDEFS",2);
			output << file->GetLump("SIDEDEFS",2);
			output << file->GetLump("VERTEXES",2);
			output << file->GetLump("SEGS",2);
			output << file->GetLump("SSECTORS",2);
			output << file->GetLump("NODES",2);
			output << file->GetLump("SECTORS",2);
			output << file->GetLump("REJECT",2);
			output << file->GetLump("BLOCKMAP",2);
		}

		// Skys
		if(i == COMBINE)
		{
			file->RenameLump("RSKY1", "MSKY1");
			skys << file->GetLump("MSKY1");
		}
		else if(i == GERYON)
		{
			file->RenameLump("RSKY1", "MSKY2");
			skys << file->GetLump("MSKY2");
		}
		else if(i == MANOR)
		{
			file->RenameLump("STARS", "MSKY3");
			skys << file->GetLump("MSKY3");
		}

		delete file;
	}

	// Merge skys
	unsigned char empty[1] = { 0 };
	output.AddLump("TX_START", 0, empty);
	output << skys.GetLump("MSKY1");
	output << skys.GetLump("MSKY2");
	output << skys.GetLump("MSKY3");
	output.AddLump("TX_END", 0, empty);

	QString mapinfoData = mapinfo;
	for(unsigned int i = 0;i < 22;++i)
		mapinfoData = mapinfoData.arg(MapProgressions[progression][i]);
	for(unsigned int i = 0;i < 21;++i)
		mapinfoData = mapinfoData.arg(MapMusic[musicSelections][i]);
	if(progression == Patcher::MP_XASERTEXT)
		mapinfoData = mapinfoData.arg(11).arg(12).arg(13).arg(xaserinfo).arg("SKY3");
	else
		mapinfoData = mapinfoData.arg(11).arg(11).arg(11).arg(clusterinfo).arg("SKY1");

	output.AddLump("MAPINFO", mapinfoData.length(), (const unsigned char*)mapinfoData.toAscii().constData());
	output.AddLump("LANGUAGE", sizeof(language), language);

	if(musicSelections == Patcher::MS_XASERPICKS || musicSelections == Patcher::MS_BLZUT3PICKS)
	{
		WadFile *extraFile = WadFile::FromFile(doom2Wad.toAscii());
		extraFile->RenameLump("D_DM2INT", "D_INTDM2");
		output << extraFile->GetLump("D_INTDM2");
		if(musicSelections == Patcher::MS_XASERPICKS)
		{
			extraFile->RenameLump("D_READ_M", "D_README");
			output << extraFile->GetLump("D_README");
		}
		delete extraFile;
	}

	if(musicSelections == Patcher::MS_XASERPICKS)
	{
		WadFile *extraFile = WadFile::FromFile(doomWad.toAscii());
		output << extraFile->GetLump("D_E1M5");
		output << extraFile->GetLump("D_E1M7");
		output << extraFile->GetLump("D_E1M8");
		output << extraFile->GetLump("D_E2M2");
		output << extraFile->GetLump("D_E2M6");
		output << extraFile->GetLump("D_E3M3");
		output << extraFile->GetLump("D_E3M8");
		delete extraFile;
	}

	//output.Dump();
	if(!output.Save(out))
		return QString("Could not write to \"") + out + "\"";
	return "";
};
